﻿//Copyright (c) Victor Hurdugaci
//http://victorhurdugaci.com
//This code is licensed under the Non-Profit Open Software License version 3.0
//Full license text: http://www.opensource.org/licenses/NOSL3.0.html

using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Interop;
using System.Runtime.InteropServices;

namespace TaskSwitcher
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class PreviewWindow : Window
    {

        #region dwmapi.dll imports

        [DllImport("dwmapi.dll", PreserveSig = false)]
        private static extern bool DwmIsCompositionEnabled();

        [DllImport("dwmapi.dll")]
        private static extern int DwmExtendFrameIntoClientArea(IntPtr hWnd, ref MARGINS pMargins);

        [StructLayout(LayoutKind.Sequential)]
        struct MARGINS
        {
            public int Left;
            public int Right;
            public int Top;
            public int Bottom;

            public void ExtendToWholeClientArea()
            {
                this.Left = -1;
                this.Right = -1;
                this.Top = -1;
                this.Bottom = -1;
            }
        }

        [DllImport("dwmapi.dll")]
        private static extern int DwmRegisterThumbnail(IntPtr dest, IntPtr src, out IntPtr thumb);

        [DllImport("dwmapi.dll")]
        private static extern int DwmUnregisterThumbnail(IntPtr thumb);

        [DllImport("dwmapi.dll")]
        private static extern int DwmQueryThumbnailSourceSize(IntPtr thumb, out PSIZE size);

        [StructLayout(LayoutKind.Sequential)]
        private struct PSIZE
        {
            public int x;
            public int y;
        }

        [DllImport("dwmapi.dll")]
        private static extern int DwmUpdateThumbnailProperties(IntPtr hThumb, ref DWM_THUMBNAIL_PROPERTIES props);

        [StructLayout(LayoutKind.Sequential)]
        private struct DWM_THUMBNAIL_PROPERTIES
        {
            public int dwFlags;
            public RECT rcDestination;
            public RECT rcSource;
            public byte opacity;
            public bool fVisible;
            public bool fSourceClientAreaOnly;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct RECT
        {
            public RECT(int left, int top, int right, int bottom)
            {
                Left = left;
                Top = top;
                Right = right;
                Bottom = bottom;
            }

            public int Left;
            public int Top;
            public int Right;
            public int Bottom;
        }

        private static readonly int DWM_TNP_VISIBLE = 0x8;
        private static readonly int DWM_TNP_RECTDESTINATION = 0x1;
        private static readonly int DWM_TNP_SOURCECLIENTAREAONLY = 0x00000010;

        #endregion

        private static readonly int ThumbnailSize = 150;
        private static readonly int ThumbnailSpacing = 10;
        private static readonly int MaxThumbnails = 5;

        private static readonly int WindowTopOffset = 30;

        private WindowsEnumerator winEnum;
        private IntPtr mainWindowHandle;

        public PreviewWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            //Store the main window handle because is needed in many places
            mainWindowHandle = new WindowInteropHelper(this).Handle;

            MakeGlassEffect();

            winEnum = new WindowsEnumerator(new IntPtr[] { mainWindowHandle });
            winEnum.Refresh();
            IEnumerable<WindowInfo> windows = winEnum.Snapshot;

            this.DataContext = windows.ElementAt(0);

            this.Width = CalculateWindowWidth(windows.Count());
            this.Height = CalculateWindowHeight(windows.Count());

            DrawWindowThumbnails();
        }

        private void DrawWindowThumbnails()
        {
            int thumbnailIndex = 0;
            foreach (WindowInfo winInfo in winEnum.Snapshot)
            {
                DrawThumbnail(winInfo, thumbnailIndex);
                ++thumbnailIndex;
            }
        }

        private void DrawThumbnail(WindowInfo win, int thumbnailIndex)
        {
            IntPtr thumbnail = win.Thumbnail;
            if (thumbnail != IntPtr.Zero)
                DwmUnregisterThumbnail(thumbnail);

            int hResult = DwmRegisterThumbnail(mainWindowHandle, win.HWnd, out thumbnail);
            if (hResult == 0)
            {
                PSIZE size;
                DwmQueryThumbnailSourceSize(thumbnail, out size);

                DWM_THUMBNAIL_PROPERTIES props = new DWM_THUMBNAIL_PROPERTIES();
                props.dwFlags = DWM_TNP_VISIBLE | DWM_TNP_RECTDESTINATION | DWM_TNP_SOURCECLIENTAREAONLY;
                props.fVisible = true;
                props.fSourceClientAreaOnly = true;

                //Set the region where the live preview will be drawn
                int left = (thumbnailIndex % MaxThumbnails) * (ThumbnailSize + ThumbnailSpacing);
                int top = (int)(thumbnailIndex / MaxThumbnails) * (ThumbnailSize + ThumbnailSpacing) + WindowTopOffset;
                int right = left + ThumbnailSize;
                int bottom = top + ThumbnailSize;

                props.rcDestination = new RECT(left, top, right, bottom);

                //Center the live preview
                if (size.x < size.y)
                {
                    double ScaleFactor = ThumbnailSize / (double)size.y;
                    int scaledX = (int)(size.x * ScaleFactor);
                    int xOffset = (ThumbnailSize - scaledX) / 2;
                    props.rcDestination.Left += xOffset;
                    props.rcDestination.Right -= xOffset;
                }
                if (size.y < size.x)
                {
                    double ScaleFactor = ThumbnailSize / (double)size.x;
                    int scaledY = (int)(size.y * ScaleFactor);
                    int yOffset = (ThumbnailSize - scaledY) / 2;
                    props.rcDestination.Top += yOffset;
                    props.rcDestination.Bottom -= yOffset;
                }

                DwmUpdateThumbnailProperties(thumbnail, ref props);
            }
        }

        private int CalculateWindowWidth(int thumbnailCount)
        {
            return thumbnailCount < MaxThumbnails ?
                thumbnailCount * (ThumbnailSize + ThumbnailSpacing) :
                MaxThumbnails * (ThumbnailSize + ThumbnailSpacing);
        }

        private int CalculateWindowHeight(int thumbnailCount)
        {
            return (int)(Math.Ceiling((double)thumbnailCount / MaxThumbnails) * (ThumbnailSize + ThumbnailSpacing)) + WindowTopOffset;
        }

        private void MakeGlassEffect()
        {
            if (DwmIsCompositionEnabled())
            {
                IntPtr mainWindowPtr = new WindowInteropHelper(this).Handle;
                HwndSource mainWindowSrc = HwndSource.FromHwnd(mainWindowHandle);
                mainWindowSrc.CompositionTarget.BackgroundColor = Colors.Transparent;

                this.Background = Brushes.Transparent;

                MARGINS margins = new MARGINS();
                margins.ExtendToWholeClientArea();

                int result = DwmExtendFrameIntoClientArea(mainWindowSrc.Handle, ref margins);

                if (result < 0)
                {
                    MessageBox.Show("An error occured while extending the glass unit.");
                    Application.Current.Shutdown();
                }
            }
        }

        private void Window_MouseMove(object sender, MouseEventArgs e)
        {
            Point mousePosition =  e.GetPosition(this);
            int windowIndex = MousePositionToWindowIndex(mousePosition);

            if (windowIndex != -1 && windowIndex < winEnum.Snapshot.Count())
            {
                //Position the selection rectangle according to the hovered preview
                this.DataContext = winEnum.Snapshot.ElementAt(windowIndex);
                SelectionMarker.SetValue(Canvas.LeftProperty, (double)(windowIndex % MaxThumbnails) * (ThumbnailSize + ThumbnailSpacing));
                SelectionMarker.SetValue(Canvas.TopProperty, Math.Floor((double)(windowIndex / MaxThumbnails)) * (ThumbnailSize + ThumbnailSpacing));
            }
        }

        private int MousePositionToWindowIndex(Point mousePosition)
        {
            int index = -1;

            if (mousePosition.Y > WindowTopOffset)
            {
                int y = (int)mousePosition.Y - WindowTopOffset;

                int yCell = (int)Math.Floor((double)y / (ThumbnailSize + ThumbnailSpacing));
                int xCell = (int)Math.Floor(mousePosition.X / (ThumbnailSize + ThumbnailSpacing));

                index = yCell * MaxThumbnails + xCell;
            }

            return index;
        }
    }
}
