﻿//Copyright (c) Victor Hurdugaci
//http://victorhurdugaci.com
//This code is licensed under the Non-Profit Open Software License version 3.0
//Full license text: http://www.opensource.org/licenses/NOSL3.0.html

using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.InteropServices;

namespace TaskSwitcher
{
    /// <summary>
    /// Represents an object that is capable of taking a snapshot of the available windows at a given moment
    /// </summary>
    sealed class WindowsEnumerator
    {
        #region user32.dll imports

        //Most of the following declarations are taken from http://www.pinvoke.net

        [DllImport("user32.dll")]
        private static extern int EnumWindows(EnumWindowsCallbackDelegate callback, int lParam = 0);

        private delegate bool EnumWindowsCallbackDelegate(IntPtr hWnd, int lParam);
        
        [DllImport("user32.dll")]
        private static extern bool IsWindowVisible(IntPtr hWnd);

        [DllImport("user32.dll")]
        private static extern IntPtr GetWindow(IntPtr hWnd, uint uCmd = GW_OWNER);

        private const uint GW_OWNER = 4;
     
        [DllImport("user32.dll")]
        private static extern IntPtr GetAncestor(IntPtr hwnd, uint gaFlags = GA_ROOTOWNER);

        private const uint GA_ROOTOWNER = 3;

        [DllImport("user32.dll")]
        private static extern IntPtr GetLastActivePopup(IntPtr hWnd);

        [DllImport("user32.dll")]
        private static extern IntPtr GetShellWindow();

        [DllImport("user32.dll")]
        private static extern int GetWindowLong(IntPtr hWnd, int nIndex = GWL_STYLE);

        private const Int32 GWL_STYLE = (-16);
        private const UInt32 WS_EX_APPWINDOW = 0x40000;

        #endregion

        //This list contains the list of windows available at the moment of calling
        //the Update method. Is readonly because will not be reinstantiated 
        private readonly List<WindowInfo> windowsSnapshot = new List<WindowInfo>();

        //A list of handles that will be ignored (filtered out) when analyzing the available windows
        private IntPtr[] WindowHandlesToIgnore = null;

        /// <summary>
        /// Creates a new instance of the windows enumerator
        /// </summary>
        /// <param name="handelsToIgnore">A list of handles (hWnd) that will be ignored from enumeration</param>
        public WindowsEnumerator(IntPtr[] handelsToIgnore)
        {
            if (handelsToIgnore == null)
                WindowHandlesToIgnore = new IntPtr[] { };
            else
                WindowHandlesToIgnore = handelsToIgnore;
        }

private static bool IsWindowChainVisible(IntPtr hWnd)
{
    // Start at the root owner
    IntPtr hwndWalk = GetAncestor(hWnd);

    // Basically we try get from the parent back to that window
    IntPtr hwndTry;
    while ((hwndTry = GetLastActivePopup(hwndWalk)) != hwndTry)
    {
        if (IsWindowVisible(hwndTry)) break;
        hwndWalk = hwndTry;
    }
    return (hwndWalk == hWnd);
}

        //This function will be called for each available window
        private bool EnumWindowsCallback(IntPtr hWnd, int lParam)
        {
            bool IsDesktopWindow = (hWnd == GetShellWindow());
            bool IsVisible = IsWindowVisible(hWnd);
            bool IsChainVisible = IsWindowChainVisible(hWnd);
            bool IsInIgnoreList = (WindowHandlesToIgnore.Contains(hWnd));
            //Filteres the taskbar
            bool IsApplicationWindow = ((GetWindowLong(hWnd) & WS_EX_APPWINDOW) == WS_EX_APPWINDOW);

            if (!IsDesktopWindow && IsVisible && IsChainVisible && !IsInIgnoreList && IsApplicationWindow)
            {           
                windowsSnapshot.Add(new WindowInfo(hWnd));
            }

            return true;
        }

        /// <summary>
        /// Gets the available snapshot
        /// </summary>
        public IEnumerable<WindowInfo> Snapshot
        {
            get
            {
                return windowsSnapshot.AsEnumerable();
            }
        }

        /// <summary>
        /// Updates the snapshot with the most recent information available
        /// </summary>
        public void Refresh()
        {
            windowsSnapshot.Clear();
            EnumWindows(EnumWindowsCallback);
        }

    }
}
